module Contrackt
  class Base
    def self.required(key)
      define_prop(Props::Required.new(key))
    end

    def self.optional(key)
      define_prop(Props::Optional.new(key))
    end

    def self.discriminator(discriminator, mapping = nil)
      @discriminator = discriminator
      @mapping = mapping
      define_prop(Props::Required.new(discriminator))
    end

    def self.define_prop(prop)
      symbol_key = prop.key.to_sym
      attr_reader symbol_key
      props[symbol_key] = prop
    end

    def self.props
      @props ||= superclass == Contrackt::Base ? {} : superclass.clone_props
    end

    def self.clone_props
      props.reduce({}) { |props, kvp| props.merge(kvp[0] => kvp[1].clone) }
    end

    def self.[]
      Collection.new(self)
    end

    def self.new(json)
      klass = determine_contract_class(json)
      klass == self ? super(json) : klass.new(json)
    end

    def self.determine_contract_class(json)
      if @discriminator
        class_key = json[@discriminator]
        (@mapping && @mapping[class_key]) || Object.const_get(class_key)
      else
        self
      end
    end

    def initialize(json)
      self.class.props.each do |symbol_key, prop|
        instance_variable_set "@#{prop.key}", prop.parse(json)
      end
    end

    def to_hash
      self.class.props.reduce({}) do |hash, array|
        symbol_key, prop = array
        hash.merge(prop.hashify(send(symbol_key)))
      end
    end
  end

  class Collection
    def initialize(klass)
      @klass = klass
    end

    def new(values)
      values.map {|value| @klass.new(value)}
    end
  end
end