module Contrackt
  module Props
    class Prop
      attr_reader :key
      def initialize(key)
        unpack(key)
      end

      def parse(json)
        target = json[key]
        if @custom_parser
          @custom_parser[target]
        elsif @klass
          @klass.new(target)
        else
          target
        end
      end

      def hashify(value)
        { key => hashify_value(value) }
      end

      def with_custom_parser(&block)
        @custom_parser = block
      end

      private
      def unpack(key)
        case key
        when String
          @key = key
        when Symbol
          @key = key
        when Hash
          raise ArgumentError, "Prop can only take one key-value pair" unless key.length == 1
          @key = key.keys[0]
          @klass = key.values[0]
        else
          raise ArgumentError, "Prop can only take a string or a hash"
        end
      end

      def hashify_value(value)
        if (value.is_a? Array)
          value.map {|value| hashify_value(value)}
        elsif value.respond_to?(:to_hash)
          value.to_hash
        else
          value
        end
      end
    end

    class Required < Prop
    end

    class Optional < Prop
      def parse(json)
        json[key].nil? ? nil : super(json)
      end

      def hashify(value)
        value.nil? ? {} : super(value)
      end
    end
  end
end